@@ -1,13 +1,11 @@
 # Net::CIDR
 #
-# Copyright 2001-2009 Sam Varshavchik.
+# Copyright 2001-2012 Sam Varshavchik.
 #
 # with contributions from David Cantrell.
 #
 # This program is free software; you can redistribute it
 # and/or modify it under the same terms as Perl itself.
-#
-# $Revision: 1.19 $
 
 package Net::CIDR;
 
@@ -19,8 +17,6 @@ require Exporter;
 # use AutoLoader qw(AUTOLOAD);
 use Carp;
 
-use Math::BigInt;
-
 @ISA = qw(Exporter);
 
 # Items to export into callers namespace by default. Note: do not export
@@ -54,7 +50,7 @@ use Math::BigInt;
 	
 );
 
-$VERSION = "0.13";
+$VERSION = "0.17";
 
 1;
 
@@ -285,13 +281,10 @@ sub cidr2range {
 	my ($ip, $pfix)=($1, $2);
 
 	my $isipv6;
-	($isipv6, $ip)=_ipv6to4($ip);
 
-	my @ips= split (/\.+/, $ip);
+	my @ips=_iptoipa($ip);
 
-	grep {
-	    croak "$_, as in '$cidr', is not a byte" unless $_ >= 0 && $_ <= 255 && $_ =~ /^[0-9]+$/;
-	} @ips;
+	$isipv6=shift @ips;
 
 	croak "$pfix, as in '$cidr', does not make sense"
 	    unless $pfix >= 0 && $pfix <= ($#ips+1) * 8 && $pfix =~ /^[0-9]+$/;
@@ -419,6 +412,28 @@ sub _ipv4to6 {
     return join(":", splice (@words, 0, $ind)) . "::"
 	. join(":", @s);
 }
+
+# An IP address to an octet list.
+
+# Returns a list. First element, flag: true if it was an IPv6 flag. Remaining
+# values are octets.
+
+sub _iptoipa {
+    my $iparg=shift;
+
+    my $isipv6;
+    my $ip;
+
+    ($isipv6, $ip)=_ipv6to4($iparg);
+
+    my @ips= split (/\.+/, $ip);
+
+    grep {
+	croak "$_, in $iparg, is not a byte" unless $_ >= 0 && $_ <= 255 && $_ =~ /^[0-9]+$/;
+    } @ips;
+
+    return ($isipv6, @ips);
+}
     
 sub _h62d {
     my $h=shift;
@@ -474,39 +489,45 @@ sub _cidr2iprange {
 #
 
 sub addr2cidr {
-	my(@blocks, $address, $bitmask, $net, $octetstash) = ();
-	my($isIPv6, $addr) = _ipv6to4(shift);
-	my $bitsInAddress = ($isIPv6)?128:32;
+    my @ips=_iptoipa(shift);
 
-	# create BigInts if necessary.  Straight assignment of ints to
-	# BigInts breaks BigInt-ness hence the bizarre subtractions later.
-	if($isIPv6) {
-		$address = new Math::BigInt 0;
-		$bitmask = new Math::BigInt 0;
-		$net = new Math::BigInt 0;
-		$octetstash = new Math::BigInt 0;
-	} else { ($address, $bitmask, $net, $octetstash) = (0, 0, 0, 0); }
+    my $isipv6=shift @ips;
 
-	# convert dotted-octets into an int (or BigInt)
-	do { $address = $address << 8; $address = $address + $_ }
-		for split(/\./, $addr);
+    my $nbits;
 
-	foreach my $bits (reverse 0..$bitsInAddress) {
-		$octetstash = $octetstash - $octetstash + 255;
-		$bitmask = $bitmask - $bitmask;
-		for(1 .. $bits) { $bitmask = ($bitmask << 1) + 1; }
-		$bitmask = $bitmask << ($bitsInAddress - $bits);
-		
-		$net = $address & $bitmask;
-		push @blocks, (map { $_ =~ s/\+//; (($isIPv6)?_ipv4to6($_):$_) }
-		    join('.', reverse map {
-			my $temp = ($octetstash & $net) >> ($_ * 8);
-			$octetstash = $octetstash << 8;
-			$temp;
-		    } (0 .. $bitsInAddress / 8 - 1)
-		))[0]."/$bits";
+    if ($isipv6)
+    {
+	croak "An IPv6 address is 16 bytes long" unless $#ips == 15;
+	$nbits=128;
+    }
+    else
+    {
+	croak "An IPv4 address is 4 bytes long" unless $#ips == 3;
+	$nbits=32;
+    }
+
+    my @blocks;
+
+    foreach my $bits (reverse 0..$nbits)
+    {
+	my @ipcpy=@ips;
+
+	my $n=$bits;
+
+	while ($n < $nbits)
+	{
+	    @ipcpy[$n / 8] &= (0xFF00 >> ($n % 8));
+
+	    $n += 8;
+
+	    $n &= 0xF8;
 	}
-	return @blocks;
+
+	my $s=join(".", @ipcpy);
+
+	push @blocks, ($isipv6 ? _ipv4to6($s):$s) . "/$bits";
+    }
+    return @blocks;
 }
 
 # Address and netmask to CIDR
@@ -609,8 +630,8 @@ sub _range2cidr {
 
     return _range2cidr8($a, $b) if $#a < 0; # Least significant octet pair.
 
-    croak unless $a >= 0 && $a <= 255 && $a =~ /^[0-9]+$/;
-    croak unless $b >= 0 && $b <= 255 && $b =~ /^[0-9]+$/ && $b >= $a;
+    croak "Bad starting address\n" unless $a >= 0 && $a <= 255 && $a =~ /^[0-9]+$/;
+    croak "Bad ending address\n" unless $b >= 0 && $b <= 255 && $b =~ /^[0-9]+$/ && $b >= $a;
 
     my @c;
 
@@ -710,8 +731,8 @@ sub _range2cidr8 {
 	my $a=shift @r;
 	my $b=shift @r;
 
-	croak unless $a >= 0 && $a <= 255 && $a =~ /^[0-9]+$/;
-	croak unless $b >= 0 && $b <= 255 && $b =~ /^[0-9]+$/ && $b >= $a;
+	croak "Bad starting address\n" unless $a >= 0 && $a <= 255 && $a =~ /^[0-9]+$/;
+	croak "Bad ending address\n" unless $b >= 0 && $b <= 255 && $b =~ /^[0-9]+$/ && $b >= $a;
 
 	++$b;
 
@@ -754,8 +775,8 @@ sub _cidr2range8 {
 	my $a=shift @c;
 	my $b=shift @c;
 
-	croak unless $a >= 0 && $a <= 255 && $a =~ /^[0-9]+$/;
-	croak unless $b >= 0 && $b <= 8 && $b =~ /^[0-9]+$/;
+	croak "Bad starting address" unless $a >= 0 && $a <= 255 && $a =~ /^[0-9]+$/;
+	croak "Bad ending address" unless $b >= 0 && $b <= 8 && $b =~ /^[0-9]+$/;
 
 	my $n= 1 << (8-$b);
 
@@ -786,7 +807,7 @@ sub _ipcmp {
     my @a=split (/\./, $aa);
     my @b=split (/\./, $bb);
 
-    croak unless $#a == $#b;
+    croak "Different number of octets in IP addresses" unless $#a == $#b;
 
     while ($#a >= 0 && $a[0] == $b[0])
     {
@@ -838,19 +859,15 @@ sub cidr2octets {
 
 	$cidr =~ s/\s//g;
 
-	croak unless ($cidr =~ /(.*)\/(.*)/);
+	croak "CIDR doesn't look like a CIDR\n" unless ($cidr =~ /(.*)\/(.*)/);
 
 	my ($ip, $pfix)=($1, $2);
 
 	my $isipv6;
 
-	($isipv6, $ip)=_ipv6to4($ip);
+	my @ips=_iptoipa($ip);
 
-	my @ips= split (/\.+/, $ip);
-
-	grep {
-	    croak "$_, as in '$cidr', is not a byte" unless $_ >= 0 && $_ <= 255 && $_ =~ /^[0-9]+$/;
-	} @ips;
+	$isipv6=shift @ips;
 
 	croak "$pfix, as in '$cidr', does not make sense"
 	    unless $pfix >= 0 && $pfix <= ($#ips+1) * 8 && $pfix =~ /^[0-9]+$/;
@@ -980,7 +997,7 @@ sub cidradd {
     my @b;
 
     grep {
-	croak unless /(.*)-(.*)/;
+	croak "This doesn't look like start-end\n" unless /(.*)-(.*)/;
 	push @a, $1;
 	push @b, $2;
     } @cidr;
@@ -1148,7 +1165,7 @@ sub cidrlookup {
     my @b;
 
     grep {
-	croak unless /(.*)-(.*)/;
+	croak "This doesn't look like start-end\n" unless /(.*)-(.*)/;
 	push @a, $1;
 	push @b, $2;
     } @cidr;
@@ -1172,10 +1189,20 @@ sub cidrlookup {
 
 =head2 $ip=Net::CIDR::cidrvalidate($ip);
 
-Validate whether $ip is a valid IPv4 or IPv6 address.
+Validate whether $ip is a valid IPv4 or IPv6 address, or a CIDR.
 Returns its argument or undef.
 Spaces are removed, and IPv6 hexadecimal address are converted to lowercase.
 
+$ip with less than four octets gets filled out with additional octets, and
+the modified value gets returned. This turns "192.168/16" into a proper
+"192.168.0.0/16".
+
+If $ip contains a "/", it must be a valid CIDR, otherwise it must be a valid
+IPv4 or an IPv6 address.
+
+A technically invalid CIDR, such as "192.168.0.1/24" fails validation, returning
+undef.
+
 =cut
 
 sub cidrvalidate {
@@ -1185,6 +1212,16 @@ sub cidrvalidate {
 
     $v=lc($v);
 
+    my $suffix;
+
+    ($v, $suffix)=($1, $2) if $v =~ m@(.*)/(.*)@;
+
+    if (defined $suffix)
+    {
+	return undef unless $suffix =~ /^\d+$/ &&
+	    ($suffix eq "0" || $suffix =~ /^[123456789]/);
+    }
+
     if ($v =~ /^([0-9\.]+)$/ || $v =~ /^::ffff:([0-9\.]+)$/ ||
 	$v =~ /^:([0-9\.]+)$/)
     {
@@ -1194,14 +1231,39 @@ sub cidrvalidate {
 
 	my @o= split(/\./, $n);
 
+	while ($#o < 3)
+	{
+	    push @o, "0";
+	}
+
+	$n=join(".", @o);
+
 	return undef if $#o != 3;
 
 	foreach (@o)
 	{
-	    return if /^0./;
-	    return if $_ < 0 || $_ > 255;
+	    return undef if /^0./;
+	    return undef if $_ < 0 || $_ > 255;
 	}
-	return $v;
+
+	if ($v =~ /^::ffff/)
+	{
+	    $suffix=128 unless defined $suffix;
+
+	    return undef if $suffix < 128-32;
+
+	    $suffix -= 128-32;
+	}
+	else
+	{
+	    $suffix=32 unless defined $suffix;
+	}
+
+	foreach (addr2cidr($n))
+	{
+	    return $_ if $_ eq "$n/$suffix";
+	}
+	return undef;
     }
 
     return undef unless $v =~ /^[0-9a-f:]+$/;
@@ -1218,7 +1280,13 @@ sub cidrvalidate {
 	return undef if length ($_) > 4;
     }
 
-    return $v;
+    $suffix=128 unless defined $suffix;
+
+    foreach (addr2cidr($v))
+    {
+	return $_ if $_ eq "$v/$suffix";
+    }
+    return undef;
 }
 
 =pod
@@ -1226,7 +1294,7 @@ sub cidrvalidate {
 =head1 BUGS
 
 Garbage in, garbage out.
-Always use validate() before doing anything with untrusted input.
+Always use cidrvalidate() before doing anything with untrusted input.
 Otherwise,
 "slightly" invalid input will work (extraneous whitespace
 is generally OK),
